package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.BipackDecoder
import net.sergeych.bipack.BipackEncoder

/*
The problem with container is the following. When we encrypt it with asymmetric key,
we provide public key as the recipient. But public key alone is not effective
as it is restricted to anonymous encryption which is very slow.

We might need an alternative to specify the (sender key,recipient key) pair also. We need a good
solution for it.

 */


@Serializable
sealed class Container {

    class InvalidContainerException : Crypto2Exception("the container is invalid")

    abstract fun decryptWith(keyRing: UniversalRing): UByteArray?

    fun decryptWith(vararg decryptingKeys: DecryptingKey): UByteArray? =
        decryptWith(UniversalRing(*decryptingKeys))

    @Transient
    var decryptedData: UByteArray? = null

    val isDecrypted: Boolean get() = decryptedData != null

    @Serializable
    @SerialName("1")
    class Single(val keyTag: KeyTag, val encryptedMessage: UByteArray) : Container() {
        override fun decryptWith(keyRing: UniversalRing): UByteArray? {
            decryptedData?.let { return it }
            for (k in keyRing) {
                if (k.tag == keyTag) {
                    kotlin.runCatching { k.decrypt(encryptedMessage) }.getOrNull()?.let {
                        decryptedData = it
                        return it
                    }
                }
            }
            return null
        }
    }

    @Serializable
    @SerialName("*")
    class Multi(val encryptedKeys: List<EncryptedKey>, val encryptedMessage: UByteArray) : Container() {
        @Serializable
        class EncryptedKey(val tag: KeyTag, val cipherData: UByteArray)

        private var mainKey: SymmetricKey? = null

        override fun decryptWith(keyRing: UniversalRing): UByteArray? {
            decryptedData?.let { return it }
            for (key in keyRing) {
                val tag = key.tag
                for (encryptedKey in encryptedKeys) {
                    if (tag == encryptedKey.tag) {
                        kotlin.runCatching {
                            BipackDecoder.decode<SymmetricKey>(
                                key.decrypt(encryptedKey.cipherData).toByteArray()
                            )
                        }.getOrNull()?.let { k ->
                            if (kotlin.runCatching { decryptedData = k.decrypt(encryptedKey.cipherData) }.isFailure)
                                throw InvalidContainerException()
                            mainKey = k
                        }
                        if (decryptedData != null) return decryptedData
                    }
                }
            }
            return null
        }
    }


    val encoded: UByteArray by lazy {
        BipackEncoder.encode(this).toUByteArray()
    }

    companion object {

        class Builder internal constructor(private val plainData: UByteArray) {
            private val plainKeys = mutableListOf<EncryptingKey>()
            private val keyPairs = mutableListOf<Pair<Asymmetric.SecretKey?,Asymmetric.PublicKey>>()
            private var fillRange: IntRange? = null

            fun key(vararg keys: EncryptingKey) { plainKeys.addAll(keys) }

            fun key(vararg pairs: Pair<Asymmetric.SecretKey,Asymmetric.PublicKey>) {
                keyPairs.addAll(pairs)
            }

            fun key(vararg publicKeys: Asymmetric.PublicKey) {
                keyPairs.addAll(publicKeys.map { null to it })
            }

            @Suppress("unused")
            fun fill(range: IntRange) {
                require(range.first >= 0 ) { "range must be positive"}
                fillRange = range
            }

            fun build(): Container {
                return when( plainKeys.size + keyPairs.size ) {
                    0 -> throw IllegalArgumentException("Container needs at least one key")
                    1 -> {
                        plainKeys.firstOrNull()?.let {
                            Single(it.tag, it.encrypt(plainData, fillRange))
                        } ?: run {
                            val (sk, pk) = keyPairs.first()
                            Single(pk.tag, pk.encryptMessage(plainData,
                                senderKey = sk ?: Asymmetric.randomSecretKey(),
                                randomFill = fillRange).encoded)
                        }
                    }
                    else -> {
                        TODO("multikey")
                    }
                }
            }
        }

        fun create(plainData: UByteArray,builder: Builder.()->Unit) =
            Builder(plainData).also { it.builder() }.build()

        fun create(plainData: UByteArray, vararg keys: EncryptingKey): Container =
            create(plainData) { key(*keys) }

        fun create(plainData: UByteArray, vararg keys: Pair<Asymmetric.SecretKey,Asymmetric.PublicKey>) =
            create(plainData) { key(*keys) }

        fun decode(encoded: UByteArray): Container {
            return BipackDecoder.decode(encoded.toByteArray())
        }
    }

}
//
//
//    /**
//     * General error while unpacking and decrypting the container
//     */
//    open class Error(reason: String = "failed to process the container") : IllegalArgumentException(reason)
//
//    /**
//     * Decryption failed
//     */
//    class DecryptionError : Error("failed to decrypt the container")
//
//    /**
//     * Packed container seems to be corrupted
//     */
//    class StructureError : Error("illegal container structure")
//
//    /**
//     * Key Ids for the key that can decrypt the container
//     */
//    abstract val keyIds: Set<KeyIdentity>
//    abstract fun decrypt(key: DecryptingKey): ByteArray
//
//    abstract fun update(keyRing: IdentifiableKeyring, newData: ByteArray): ByteArray?
//
//    fun decrypt(keyRing: IdentifiableKeyring): Pair<IdentifiableKey, ByteArray>? {
//        // Fast search: look for a key that shurely should
//        for (k in keyRing.byIdAndSymmetrics(keyIds)) {
//            try {
//                if( k is DecryptingKey )
//                    return k to decrypt(k)
//            } catch (x: Throwable) {
//                // This means our key is not ok. It a rare but possible collision
//            }
//        }
////        // some symmetric key can sometimes decrypt even if their ID is wrong, so
////        // we check them all:
////        for (k in keyRing.keys) {
////            if (k is SymmetricKey) {
////                try {
////                    return k to decrypt(k)
////                } catch (x: Throwable) {
////                    // it's ok. we just tried
////                }
////            }
////        }
//        return null
//    }
//
//    @Serializable
//    @SerialName("single")
//    data class Single(val keyId: KeyIdentity, val ciphertext: ByteArray) : Container() {
//        override val keyIds by lazy { setOf(keyId) }
//        override fun decrypt(key: DecryptingKey): ByteArray = try {
//            key.etaDecrypt(ciphertext)
//        } catch (x: Exception) {
//            throw DecryptionError()
//        }
//
//        override fun update(keyRing: IdentifiableKeyring, newData: ByteArray): ByteArray? {
//            return decrypt(keyRing)?.let { (k, _) ->
//                if (k !is EncryptingKey) throw IllegalArgumentException("key is not suitable for re-encryption")
//                encryptData(newData, k)
//            }
//        }
//    }
//
//    @Serializable
//    data class EncryptedKey(val id: KeyIdentity, val encryptedKey: ByteArray)
//
//    @Serializable
//    @SerialName("multi")
//    data class Multiple(val keys: List<EncryptedKey>, val ciphertext: ByteArray) : Container() {
//        override val keyIds: Set<KeyIdentity> by lazy { keys.map { it.id }.toSet() }
//
//        override fun decrypt(key: DecryptingKey): ByteArray {
//            for (k in keys) {
//                // Simple: we just use ID
//                // Complex: wrong ID with symmetric key that can decrypt:
//                if (key is SymmetricKey || k.id == key.id) {
//                    try {
//                        val dataKey = SymmetricKeys.create(key.etaDecrypt(k.encryptedKey), k.id)
//                        return dataKey.etaDecrypt(ciphertext)
//                    } catch (x: Throwable) {
//                        // might be wrong symmetric key id
//                        if( key !is SymmetricKey) throw DecryptionError()
//                    }
//                }
//            }
//            throw IllegalArgumentException("the key does not open this container")
//        }
//
//        override fun update(keyRing: IdentifiableKeyring, newData: ByteArray): ByteArray? {
//            for (k in keys) {
//                for (key in keyRing[k.id]) {
//                    if (key is DecryptingKey) {
//                        val dataKey = SymmetricKeys.create(key.etaDecrypt(k.encryptedKey), k.id)
//                        return BossEncoder.encode(Multiple(keys, dataKey.etaEncrypt(newData)) as Container)
//                    }
//                }
//            }
//            return null
//        }
//    }
//
//
////            return decrypt(keyRing)?.let { (key,_) ->
////                if(key !is EncryptingKey) throw IllegalArgumentException("key is not suitable for re-encryption")
////                if(key !is DecryptingKey) throw IllegalArgumentException("key is not suitable for re-encryption (not decrypting!)")
////                val dataKey = SymmetricKeys.create( key.etaDecrypt(k.encryptedKey), k.id)
////                encryptData(newData,k)
////            }
////        }
////    }
//
//    companion object : Loggable by LogTag("CRCON") {
//        /**
//         * Create single-key container. Most often you should call [encrypt] instead.
//         */
//        fun single(payload: ByteArray, key: EncryptingKey): ByteArray {
//            return BossEncoder.encode(Single(key.id, key.etaEncrypt(payload)) as Container)
//        }
//
//        /**
//         * Create nultiple keys container event with one key. Most often you should not call it but
//         * use [encrypt] instead.
//         */
//        fun multi(payload: ByteArray, keys: List<EncryptingKey>): ByteArray {
//            val dataKey = SymmetricKeys.random()
//            return BossEncoder.encode(
//                Multiple(
//                    keys.map { EncryptedKey(it.id, it.etaEncrypt(dataKey.packed)) },
//                    dataKey.etaEncrypt(payload)
//                ) as Container
//            )
//        }
//
//        /**
//         * Preferred method to create encrypted container. Selects the type automatically. Decrypt it with
//         * any of the provided key (their [DecryptingKey] counterparts) with [decrypt]
//         *
//         * @param payload data to encrypt. Could be `String`, `ByteArray` or any serializable type. Can't be null.
//         * @param keys one or more keys to encrypt with.
//         * @return encryped and packed container
//         */
//        inline fun <reified T> encrypt(payload: T, vararg keys: EncryptingKey): ByteArray =
//            encryptData(BossEncoder.encode(payload), *keys)
//
//        /**
//         * Encrypt payload  as is, without boss serialization. See [encrypt] for details.
//         * @return encrypted packed container
//         */
//        fun encryptData(payload: ByteArray, vararg keys: EncryptingKey): ByteArray =
//            when (keys.size) {
//                0 -> throw IllegalArgumentException("provide at least one key")
//                1 -> single(payload, keys[0])
//                else -> multi(payload, keys.toList())
//            }
//
//        /**
//         * Update packed container using proper key in the ring with a new payload. This function os useful
//         * when the container could be multiple (but it works with any) and we know only one of its keys,
//         * but want to change the data. It is possible as any of the keys could be used to re-encrypt data
//         * that _will be available to all other keys_.
//         * @param payload packed container
//         * @param keyRing keyring which should contain at least one key that can decrypt the contatiner
//         * @param packed new payload to encrypt
//         * @return packed encrypted container with new payload or null if keyring can't decrypt it
//         */
//        inline fun <reified T> update(packed: ByteArray, keyRing: IdentifiableKeyring, payload: T): ByteArray? {
//            return packed.decodeBoss<Container>().update(keyRing, BossEncoder.encode(payload))
//        }
//
//        /**
//         * Try to decrypt a container using some keys.
//         *
//         * @param keys to try to open the container.
//         * @return the decrypted paylaod on success or null if no keys could open it.
//         */
//        inline fun <reified T> decrypt(packed: ByteArray, vararg keys: DecryptingKey): T? =
//            decrypt<T>(packed, Keyring(*keys))
//
//        /**
//         * decrypt the container using leys in a keyring.
//         * @return recrypted payload or null if no key from a keyring can open it.
//         */
//        inline fun <reified T> decrypt(packed: ByteArray, ring: IdentifiableKeyring): T? {
//            return decryptAsBytes(packed, ring)?.decodeBoss<T>()
//        }
//
//        fun decryptAsBytes(packed: ByteArray, ring: IdentifiableKeyring): ByteArray? =
//            protect {
//                packed.decodeBoss<Container>().decrypt(ring)?.second
//            }
//
//        fun decryptAsBytes(packed: ByteArray, vararg keys: DecryptingKey): ByteArray? =
//            protect {
//                packed.decodeBoss<Container>().decrypt(Keyring(*keys))?.second
//            }
//
//        private inline fun <reified T> protect(f: () -> T): T =
//            try {
//                f()
//            } catch (t: Throwable) {
//                throw StructureError()
//            }
//    }
//}
